%******************************************************************************;
%* SAS macro MPR2SR   vs. 0.0.1 , 28 Jun 2001, by Jim Groeneveld, Netherlands *;
%******************************************************************************;
%*                                                                            *;
   %MACRO   MPR2SR     /* Convert Multiple Partial Records to Single Records  */
/* -------------------------------------------------------------------------- */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
(  SourceDs=_LAST_     /* Source data set specification                     R */
,  TargetDs=           /* Target data set specification (may be SourceDs)   R */
,  OverWrit=N          /* specify Y to overwrite an existing TargetDs       O */
,  IdList  =           /* Variable List of which multiple Identical value   R */
                       /*     combinations determine a single new record      */
                       /*     (often variables like a Case Number)            */
,  IdFormat=UNFORMATTED/* identify cases by unformatted values of identifi- O */
                       /*     cation variables, alternative argument value:   */
                       /*   FORMATTED (or explicitely empty) : identify       */
                       /*     cases by eventually formatted values of         */
                       /*     identification variables                        */
,  MRList  =           /* Multiple Records variable name List to keep,      O */
                       /*     No value or '_ALL_' involves all variables      */
,  Index   =           /* Indexing measures variable name: numeric integer  O */
                       /*     and unique within IdList, default automatic ~)  */
,  Lowest  =           /*  Lowest (integer) value of index variable         O */
,  Highest =           /* Highest (integer) value of index variable         O */
,  KeepIndx=Y          /* keep (redundant) Index variable from last record  O */
,  KeepDups=N          /* keep duplicate records (equal index values): Y/N  O */
,  KeepRest=N          /* keep one unspecified variables value anyway: Y/N  O */
,  ForMyUse=           /* argument for the author's use only, do not use it ! */
)/ DES     =             'Multiple Partial Records-Single Records' /* STORE */ ;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* File       : MPR2SR.SAS / MPR2SR.040 ( / MPR2SR__.SAS )                    *;
%* Directory  : .................                                             *;
%*                                                                            *;
%* Purpose    : Transformation of datasets with exclusive partial records of  *;
%*              non-missing variables in multiple records (a record per part) *;
%*              into datasets with all parts within a single record, or:      *;
%*              Transformation of datasets with both master and transaction   *;
%*              records into datasets with single updated records.            *;
%*                                                                            *;
%* Programmer : Jim Groeneveld                                                *;
%* Date       : 28 Jun 2001                                                   *;
%* Version    : 0.40                                                          *;
%*                                                                            *;
%* SAS version: 6.12 under Windows 95                                         *;
%*                                                                            *;
%* Validation : -                                                             *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* Notes:                                                                     *;
%* ======                                                                     *;
%* An alternative algorithm would be to process all records one by one, while *;
%* remembering their (sorted according to IdList and Index variables)         *;
%* immediately preceding record (in temporary array variables of all types),  *;
%* checking for missing values to be replaced by values from the temporary    *;
%* memory. This would need quite a lot of programming code and has the        *;
%* advantage to recognize non-unique, non-missing values of variables.        *;
%*                                                                            *;
%* The algorithm applied here is a much better and faster one, leaving as     *;
%* much as possible to SAS to do. It basically uses the UPDATE data step      *;
%* statement, involving a master and transaction dataset. After generating an *;
%* empty master dataset with all variables, the original dataset (when sorted)*;
%* can be used as the transaction dataset with multiple transaction records.  *;
%* This algorithm uses very little (only a few lines) really essential code   *;
%* and probably is faster than the alternative algorithm.                     *;
%*                                                                            *;
%* In case there are more than one records with the same index value for any  *;
/* combination of IdList variable's values, there would be multiple identical */
%* transaction records for that IdList case, each of which are being          *;
%* processed. If only one (the first encountered) transaction record should   *;
%* be used, duplicates must be avoided by the argument KeepDups=N (default).  *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* History of MPR2SR:                                                         *;
%* ==================                                                         *;
%* version  date         description of changes, additions, improvements      *;
%* -------  -----------  ---------------------------------------------------- *;
%* 0.0      13 Jun 2001  1st working release converting many data structures  *;
%*                       (effect of formatted IdList still should be checked) *;
%* 0.1      18 Jun 2001  added MRList, Check, KeepIndx and KeepRest arguments,*;
%*                       added condition: check on single non-missing per var *;
%* 0.2      19 Jun 2001  added: check on single for each IdList combination   *;
%* 0.3      23 Jun 2001  added: DupAsOne argument for interpreting duplicates *;
%* 0.4      28 Jun 2001  split off Chk4EP doing the check for MPR structure   *;
%*                                                                            *;
%******************************************************************************;

%* OPTIONS MPRINT MERROR SERROR MLOGIC SYMBOLGEN MACROGEN;
%* remove comment chars for testing ;

  %IF ( %UPCASE(&ForMyUse) = COMPILE ) %THEN
  %DO;
    %*ValueLst (ForMyUse=Compile);
    %GOTO Finish;
  %END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* LOCAL macro variables declarations;
%*___________________________________;

 %LOCAL    /* Define internal macro variables explicitely local */
  MacName  /* This macros external name: MPR2SR */
  I        /* element index */
  Id       /* IdList element */
  ByList   /* all IdList elements eventually preceeded by "GROUPFORMAT" */
  IdFrqLst /* IdList with elements separated by singele *'s indicating tables */
  LastId   /* last Id from IdList */
  MultRec  /* MRlist element */
  _Sorted1 /* stand-in name for sorted dataset 1 XX_Sort1 */
  _Sorted2 /* stand-in name for sorted dataset 2 XX_Sort2 */
  _Extreme /* stand-in name for Extremes dataset XX_Extr_ */
  Contents /* stand-in name for Contents dataset XX_Cont_ */
  Scratch  /* stand-in name for Scratch datasets XX_Scrat */
  ValList  /* temporary variable value list storage */
  VarValue /* temporary variable value storage, element of ValList */
  Sequence /* temporary variable sequence number storage */
  ErrCount /* number of erroneously specified arguments */
  Err      /* text 'E R R O R' as consecutive characters */
  Warn     /* text 'W A R N I N G' as consecutive characters */
  ;

%* Stand-in variable and dataset names;
  %LET MacName  = MPR2SR  ;
  %LET _Sorted1 = XX_Sort1;
  %LET _Sorted2 = XX_Sort2;
  %LET _Extreme = XX_Extr_;
  %LET Contents = XX_Cont_;
  %LET Scratch  = XX_Scrat;
  %LET ValList  = XX_VaLst;
  %LET VarValue = XX_Value;
  %LET Sequence = XX_SeqNr;
  %LET Err      = %SYSFUNC(COMPRESS(E R R O R, %STR( )));
  %LET Warn     = %SYSFUNC(COMPRESS(W A R N I N G, %STR( )));

%* Macro header in LOG file;
  %PUT;
  %PUT |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~|;
  %PUT |   Macro &MacName, vs. 0.4, by Jim Groeneveld, 28 Jun 2001   |;
  %PUT |   Multiple Partial Records To Single Records Conversion   |;
  %PUT |___________________________________________________________|;
  %PUT;

%* Feedback of specified arguments;
  %PUT &MacName: SourceDs = &SourceDs;
  %PUT &MacName: TargetDs = &TargetDs;
  %PUT &MacName: OverWrit = &OverWrit;
  %PUT &MacName: IdList   = &IdList  ;
  %PUT &MacName: IdFormat = &IdFormat;
  %PUT &MacName: MRlist   = &MRlist  ;
  %PUT &MacName: Index    = &Index   ;
  %PUT &MacName: Lowest   = &Lowest  ;
  %PUT &MacName: Highest  = &Highest ;
  %PUT &MacName: KeepIndx = &KeepIndx;
  %PUT &MacName: KeepDups = &KeepDups;
  %PUT &MacName: KeepRest = &KeepRest;
  %*PUT &MacName: ForMyUse = &ForMyUse;

%* Convert lower case letters to upper case and keep first argument characters;
  %IF ( &OverWrit NE ) %THEN %LET OverWrit = %UPCASE(%SUBSTR(&OverWrit,1,1));
  %IF ( &KeepIndx NE ) %THEN %LET KeepIndx = %UPCASE(%SUBSTR(&KeepIndx,1,1));
  %IF ( &KeepDups NE ) %THEN %LET KeepDups = %UPCASE(%SUBSTR(&KeepDups,1,1));
  %IF ( &KeepRest NE ) %THEN %LET KeepRest = %UPCASE(%SUBSTR(&KeepRest,1,1));

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
* Check arguments for legal contents and discrepancies, calling e_r_r_o_r_s    ;
*______________________________________________________________________________;

  %LET ErrCount = 0;
  %IF ( &Lowest NE AND &Highest NE ) %THEN
  %DO;
    %IF ( &Lowest > &Highest ) %THEN
    %DO;
      %PUT *** &MacName *** &ERR: Lowest (&Lowest) larger than Highest (&Highest),;
      %PUT                       macro &MacName will abort;
      %LET ErrCount = %EVAL ( &ErrCount + 1);
    %END;
  %END;
  %IF ( &SourceDs EQ ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: SourceDs argument not specified, macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE %IF ( %SYSFUNC(EXIST(&SourceDs)) EQ 0 ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Dataset &SourceDs does not exist,;
    %PUT                       macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE
  %DO; %* Check for non-existing variables within dataset;
   %IF ( %UPCASE(&MRlist) NE %STR(_ALL_) AND &MRlist NE ) %THEN
    %ChkVar (Data=&SourceDs, VarList=&IdList &MRlist &Index, CalledBy=&MacName,
             Result = ErrCount); %* adding up non-existent variables ;
   %ELSE
    %ChkVar (Data=&SourceDs, VarList=&IdList &Index, CalledBy=&MacName,
             Result = ErrCount); %* adding up non-existent variables ;
  %END;
  %IF ( &TargetDs EQ ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: TargetDs argument not specified, macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE %IF ( %SYSFUNC(EXIST(&TargetDs)) EQ 1 AND &OverWrit NE Y ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Dataset &TargetDs exists and OverWrit is not Y,;
    %PUT                       macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %IF ( &IdList EQ ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: IdList argument not specified, macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
/*=========
  %IF ( &MRlist EQ ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: MRlist argument not specified, macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
=========*/
  %IF ( &ErrCount GT 0 ) %THEN %GOTO Interupt;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Determine last Id variable from IdList and build BY list with "GROUPFORMAT" ;
%*_____________________________________________________________________________;

  %IF ( %UPCASE(&IdFormat) NE UNFORMATTED ) %THEN %LET ByList = ;
  %LET I = 1;
  %LET Id = %SCAN(&IdList,&I,%STR( ));
  %DO %WHILE (&Id NE);
    %LET LastId = &Id;
    %IF ( %UPCASE(&IdFormat) NE UNFORMATTED ) %THEN
      %LET ByList = &ByList GROUPFORMAT &Id;
    %LET I = %EVAL( &I + 1 );
    %LET Id = %SCAN(&IdList,&I,%STR( ));
  %END; %*=== While;

  %IF ( %UPCASE(&IdFormat) EQ UNFORMATTED ) %THEN %LET ByList = &IdList;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Check on numeric index variable ;
%*_________________________________;

%IF ( &Index NE ) %THEN
%DO;
  PROC CONTENTS DATA=&SourceDs (KEEP=&Index)
                 OUT=&Contents (KEEP=Name Type) NOPRINT;
  RUN;

  DATA _NULL_;
    SET &Contents;
* Single record with &Index, check its type;
    IF ( Type NE 1 ) THEN
    DO;
      PUT "*** &MacName *** &ERR: Index variable &Index is not numeric,";
      PUT "                      macro &MacName will abort";
      CALL SYMPUT ("ErrCount", "1");
    END;
  RUN;                  %* @@@@@@@@@ Why not character index variable? @@@@@@@@@;
%END;

PROC DATASETS NOLIST;
  DELETE &Contents;                        %* Delete scratch dataset &Contents;
RUN;

%IF ( &ErrCount GT 0 ) %THEN %GOTO Interupt;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Automatic generation of index variable (if none specified) ;
%*____________________________________________________________;

%IF ( &Index EQ ) %THEN
%DO;
  %LET Index = XX_Index; %* XX_Index assumed unique temporary name;
  PROC SORT DATA=&SourceDs
   %IF (&KeepRest EQ N AND %UPCASE(&MRlist) NE _ALL_ AND %UPCASE(&MRlist) NE )
   %THEN
   %DO;
    (KEEP=&IdList &MRlist)
   %END;
    OUT=&_Sorted1; BY &IdList;   %* Sort input file;
   %IF ( %UPCASE(&IdFormat) EQ UNFORMATTED ) %THEN %STR(FORMAT &IdList;) ;
  RUN;
  DATA &_Sorted1;
    SET &_Sorted1; BY &ByList;
    IF ( First.&LastId ) THEN &Index = %IF ( &Lowest NE ) %THEN &Lowest; %ELSE 1;;
    ELSE &Index + 1;
  RUN;
  %LET SourceDs = &_Sorted1; %* let SourceDs point to the resulting dataset;
  %LET KeepIndx = N; %* Force dropping temporary generated Index from TargetDs;
%END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Automatic determination of minimum and maximum Index values ;
%*_____________________________________________________________;

%IF ( &Lowest EQ OR &Highest EQ ) %THEN
%DO;
  PROC UNIVARIATE DATA=&SourceDs (KEEP=&Index) NOPRINT;
    VAR &Index;
    OUTPUT OUT = &_Extreme
           MIN = Minimum
           MAX = Maximum ;
  RUN;

  DATA _NULL_;
    SET &_Extreme;
    %IF ( &Lowest  EQ ) %THEN CALL SYMPUT ("Lowest" , TRIM(LEFT(PUT(Minimum,9.0))));;
    %IF ( &Highest EQ ) %THEN CALL SYMPUT ("Highest", TRIM(LEFT(PUT(Maximum,9.0))));;
  RUN;
%END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Convert eventual real Lowest and Highest values to integers ;
%*_____________________________________________________________;

 %IF (&Lowest  NE .) %THEN  %LET Lowest   = %SYSEVALF(&Lowest ,CEIL);
 %IF (&Highest NE .) %THEN  %LET Highest  = %SYSEVALF(&Highest,CEIL);

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Check lowest and highest values ;
%*_________________________________;

%IF ( &Lowest > &Highest ) %THEN
%DO;
  %PUT *** &MacName *** &ERR: Lowest (&Lowest) larger than Highest (&Highest),;
  %PUT                      macro &MacName will abort;
  %GOTO Aborting;
%END;
%ELSE %PUT Lowest index value=&Lowest and Highest index value=&Highest;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Macro actually starts here ;
%*____________________________;

%* Presort dataset according to &IdList and &Index, argument dependent;
PROC SORT DATA=&SourceDs
  %IF (&KeepRest EQ N AND %UPCASE(&MRlist) NE _ALL_ AND %UPCASE(&MRlist) NE )
  %THEN
  %DO;
    (KEEP=&IdList &MRlist &Index)
  %END;
  %IF ( &KeepDups=N ) %THEN
  %DO;
    NODUPKEY
  %END;
    OUT=&_Sorted2; BY &IdList &Index;  %* Sort input file;
  %IF ( %UPCASE(&IdFormat) EQ UNFORMATTED ) %THEN %STR(FORMAT &IdList;) ;
  FORMAT &Index;             %* Dissociate eventual format from index variable;
RUN;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Remove redundant multiple record space by UPDATing to single records;
%*_____________________________________________________________________;

* Preselect records satisfying conditions and drop the index variable;
DATA &_Sorted2;
  SET &_Sorted2;
  IF NOT ( &Index LE .z OR ( &Index GE &Lowest AND &Index LE &Highest ) ) THEN
  DO;
    PUT "*** &MacName *** &WARN: &Index value " &Index
        "out of range &Lowest...&Highest for &IdList value(s) " &IdList;
    DELETE; * such a record;
  END;
RUN;

* Create empty master dataset;
DATA &TargetDs; SET &_Sorted2; STOP; RUN;

* Update master with (multiple) transaction records from SourceDs all at once;
DATA &TargetDs %IF (&KeepIndx EQ N) %THEN (DROP=&Index);;
  UPDATE &TargetDs &_Sorted2; BY &Bylist; * This is the core of the macro;
RUN;

%Aborting:

PROC DATASETS NOLIST;
  DELETE &_Sorted1 &_Sorted2 &_Extreme;   %* Delete scratch datasets;
RUN;

%Interupt:

 %IF ( &ErrCount GT 0 ) %THEN
 %DO;
   DATA _NULL_;
     ABORT /*ABEND*/;
   RUN;
 %END;

%Finish:

%***********;
%MEND MPR2SR; * --------------------end-of-macro-MPR2SR------------------------;
%***********;

